/*
 * rsvg-show (forked from test-display)
 *
 * Copyright (C) 2002-2004 Dom Lachowicz (test-display)
 * Copyright (C) 2021 Tibor 'Igor2' Palinkas (rsvg-show)
 *
 * This program is released into the PUBLIC DOMAIN, and is meant to be a
 * useful example if how to draw a SVG image inside of a GtkWidget. This 
 * program is free software; you can redistribute it and/or modify it at your
 * will.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE
 *
 * Simple utility to view a SVG file inside of a GtkWindow
 */

#include "rsvg.h"

/* don't do internationalization */
#define N_(x) x
#define _(x) x

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <locale.h>
#include <math.h>

#include <gtk/gtk.h>
#include <gdk/gdk.h>


#if !GLIB_CHECK_VERSION (2, 36, 0)
#  define RSVG_G_TYPE_INIT g_type_init ()
#else
#  define RSVG_G_TYPE_INIT {}
#endif

#include <librsvg/rsvg.h>

#define RSVG_TYPE_IMAGE (rsvg_image_get_type ())
#define RSVG_IMAGE(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), RSVG_TYPE_IMAGE, RsvgImage))

typedef struct _RsvgImage RsvgImage;
typedef struct _RsvgImageClass RsvgImageClass;

struct _RsvgImage {
	GtkWidget parent_instance;

	cairo_surface_t *surface;	/* a cairo image surface */
};

struct _RsvgImageClass {
	GtkWidgetClass parent_class;
};

static long reload_ctr = 0;
static GType rsvg_image_get_type(void);
static cairo_surface_t *load(int reload);

static void rsvg_image_take_surface(RsvgImage * image, cairo_surface_t * surface)
{
	if (image->surface == surface)
		return;
	if (image->surface)
		cairo_surface_destroy(image->surface);
	image->surface = surface;	/* adopted */

	gtk_widget_queue_resize(GTK_WIDGET(image));
}

G_DEFINE_TYPE(RsvgImage, rsvg_image, GTK_TYPE_WIDGET);

static void rsvg_image_init(RsvgImage * image)
{
	gtk_widget_set_has_window(GTK_WIDGET(image), FALSE);
}

static void rsvg_image_finalize(GObject * object)
{
	RsvgImage *image = RSVG_IMAGE(object);

	rsvg_image_take_surface(image, NULL);
}

static void rsvg_image_get_preferred_width(GtkWidget * widget, gint * minimum, gint * natural)
{
	*minimum = *natural =  1;
}

static void rsvg_image_get_preferred_height(GtkWidget * widget, gint * minimum, gint * natural)
{
	*minimum = *natural = 1;
}

static gboolean rsvg_image_draw(GtkWidget * widget, cairo_t * cr)
{
	RsvgImage *image = RSVG_IMAGE(widget);

	if (image->surface == NULL)
		return FALSE;

	cairo_save(cr);
	cairo_set_source_surface(cr, image->surface, 0, 0);
	cairo_paint(cr);
	cairo_restore(cr);

	return FALSE;
}

static void rsvg_image_class_init(RsvgImageClass * klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS(klass);

	gobject_class->finalize = rsvg_image_finalize;
	widget_class->get_preferred_width = rsvg_image_get_preferred_width;
	widget_class->get_preferred_height = rsvg_image_get_preferred_height;
	widget_class->draw = rsvg_image_draw;
}

static RsvgImage *rsvg_image_new_take_surface(cairo_surface_t * surface)
{
	RsvgImage *image;

	image = g_object_new(RSVG_TYPE_IMAGE, NULL);
	rsvg_image_take_surface(image, surface);

	return image;
}

static cairo_surface_t *rsvg_image_get_surface(RsvgImage * image)
{
	return image->surface;
}

/* Main */

static char *_rsvg_basename(const char *file)
{
	if (file && *file)
		return g_path_get_basename(file);

	return NULL;
}

typedef struct _ViewerCbInfo ViewerCbInfo;
struct _ViewerCbInfo {
	GtkWidget *window;
	GtkWidget *popup_menu;
	RsvgImage *image;
	RsvgHandle *handle;
	GtkAccelGroup *accel_group;
	char *base_uri;
	char *id;
	RsvgDimensionData dimensions;
	gdouble x_zoom, y_zoom, x_pan, y_pan;
	int image_valid;
	int last_width, last_height;
	GtkWidget *drawing_host;
	cairo_surface_t *last_surface;
	GdkRGBA bg_color;

	/* pan UI state */
	gdouble pan_aox, pan_aoy, pan_mox, pan_moy;
	int panning;
	int mouse_x, mouse_y;
};


static cairo_surface_t *render_to_surface(ViewerCbInfo * info)
{
	int width, height;
	cairo_matrix_t matrix;
	cairo_surface_t *surface;
	cairo_t *cr;
	GtkAllocation alloc;

	if (info->image_valid) {
		gtk_widget_get_allocation(GTK_WIDGET(info->image), &alloc);
		width = alloc.width;
		height = alloc.height;
		if ((width == 1) && (height == 1)) { /* during reload the widget is destroyed so size is unknown */
			width = info->last_width;
			height = info->last_height;
		}
		else {
			info->last_width = width;
			info->last_height = height;
		}
	}
	else {
		width = ceil((double)info->dimensions.width * info->x_zoom);
		height = ceil((double)info->dimensions.height * info->y_zoom);
	}

	surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, width, height);
	if (cairo_surface_status(surface) != CAIRO_STATUS_SUCCESS) {
		cairo_surface_destroy(surface);
		return NULL;
	}

	cr = cairo_create(surface);

	cairo_matrix_init_identity(&matrix);
	cairo_matrix_translate(&matrix, info->x_pan, info->y_pan);
	cairo_matrix_scale(&matrix, info->x_zoom, info->y_zoom);
	cairo_transform(cr, &matrix);

	if (!rsvg_handle_render_cairo_sub(info->handle, cr, info->id)) {
		cairo_destroy(cr);
		cairo_surface_destroy(surface);
		return NULL;
	}

	cairo_destroy(cr);

	if (cairo_surface_status(surface) != CAIRO_STATUS_SUCCESS) {
		g_printerr("Error while rendering image: %d\n", cairo_surface_status(surface));
		cairo_surface_destroy(surface);
		return NULL;
	}

	return surface;
}

static void set_window_title(ViewerCbInfo * info)
{
	char *title;
	gchar *zoom_string;

	if (info->x_zoom != info->y_zoom) {
		zoom_string = g_strdup_printf("%4d%% : %4d%% [%ld]", (gint) (info->x_zoom * 100), (gint) (info->y_zoom * 100), reload_ctr);
	}
	else {
		zoom_string = g_strdup_printf("%4d%% [%ld]", (gint) (info->x_zoom * 100), reload_ctr);
	}

	if (info->id) {
		title = g_strdup_printf("%s#%s (%s) — %s", info->base_uri, info->id, zoom_string, _("SVG Viewer"));
	}
	else {
		title = g_strdup_printf("%s (%s) — %s", info->base_uri, zoom_string, _("SVG Viewer"));
	}
	gtk_window_set_title(GTK_WINDOW(info->window), title);
	g_free(title);
	g_free(zoom_string);
}

#define ATX(info, X) (((X) - (info)->x_pan) / (info)->x_zoom)
#define ATY(info, Y) (((Y) - (info)->y_pan) / (info)->y_zoom)

static void zoom_image(ViewerCbInfo * info, gdouble factor)
{
	double at_x = ATX(info, info->mouse_x), at_y = ATY(info, info->mouse_y); /* remember mouse cursor pos in svg coords */

	info->x_zoom *= factor;
	info->y_zoom *= factor;

	/* restore panning so that the spot under the mouse cursor has the same svg coords */
	info->x_pan = info->mouse_x - at_x * info->x_zoom;
	info->y_pan = info->mouse_y - at_y * info->y_zoom;

	rsvg_image_take_surface(info->image, render_to_surface(info));

	set_window_title(info);
}

static void pan_image(ViewerCbInfo * info, gdouble dx, gdouble dy)
{
	info->x_pan += dx;
	info->y_pan += dy;

	rsvg_image_take_surface(info->image, render_to_surface(info));
}

static void zoom_in(GObject * ignored, ViewerCbInfo * info)
{
	zoom_image(info, sqrt(G_SQRT2));
}

static void zoom_out(GObject * ignored, ViewerCbInfo * info)
{
	zoom_image(info, 1. / sqrt(G_SQRT2));
}


static void center(GObject * ignored, ViewerCbInfo * info)
{
	double ix = (double)info->dimensions.width * info->x_zoom;
	double iy = (double)info->dimensions.height * info->y_zoom;

	info->x_pan = ((double)info->last_width - ix) / 2;
	info->y_pan = ((double)info->last_height - iy) / 2;
	rsvg_image_take_surface(info->image, render_to_surface(info));
}

static void reset_zoom(GObject * ignored, ViewerCbInfo * info)
{
	info->x_pan = info->y_pan = 0;
	info->x_zoom = info->y_zoom = 1.0;

	rsvg_image_take_surface(info->image, render_to_surface(info));
}

static void zoom_extents(GObject * ignored, ViewerCbInfo * info)
{
	double ix = info->dimensions.width, iy = info->dimensions.height;
	double zx, zy;

	zx = (double)info->last_width / ix;
	zy = (double)info->last_height / iy;

	info->x_zoom = info->y_zoom = MIN(zx, zy);

	center(ignored, info);
}

static void invert_bg_color(GObject * ignored, ViewerCbInfo * info)
{
	GtkWidget *parent_widget = gtk_widget_get_parent(GTK_WIDGET(info->image));
	info->bg_color.red = 1.0 - info->bg_color.red;
	info->bg_color.green = 1.0 - info->bg_color.green;
	info->bg_color.blue = 1.0 - info->bg_color.blue;
	gtk_widget_override_background_color(parent_widget, GTK_STATE_FLAG_NORMAL, &info->bg_color);
}

static void begin_print(GtkPrintOperation * operation, GtkPrintContext * context, gpointer user_data)
{
	gtk_print_operation_set_n_pages(operation, 1);
}

static void draw_page(GtkPrintOperation * operation, GtkPrintContext * context, gint page_nr, gpointer user_data)
{
	ViewerCbInfo *info = (ViewerCbInfo *) user_data;
	cairo_t *cr;
	gdouble page_width, page_height, page_aspect;
	gdouble width, height, aspect;
	cairo_matrix_t matrix;

	cr = gtk_print_context_get_cairo_context(context);
	page_width = gtk_print_context_get_width(context);
	page_height = gtk_print_context_get_height(context);
	page_aspect = page_width / page_height;

	// FIXMEchpe
	rsvg_handle_set_dpi_x_y(info->handle, gtk_print_context_get_dpi_x(context), gtk_print_context_get_dpi_y(context));

	width = info->dimensions.width;
	height = info->dimensions.height;
	aspect = width / height;

	if (aspect <= page_aspect) {
		width = page_height * aspect;
		height = page_height;
	}
	else {
		width = page_width;
		height = page_width / aspect;
	}

	cairo_save(cr);
	cairo_matrix_init_scale(&matrix, width / info->dimensions.width, height / info->dimensions.height);
	cairo_transform(cr, &matrix);
	rsvg_handle_render_cairo(info->handle, cr);
	cairo_restore(cr);
}

static void print_pixbuf(GObject * ignored, gpointer user_data)
{
	GtkPrintOperation *print;
	ViewerCbInfo *info = (ViewerCbInfo *) user_data;

	print = gtk_print_operation_new();

	g_signal_connect(print, "begin_print", G_CALLBACK(begin_print), info);
	g_signal_connect(print, "draw_page", G_CALLBACK(draw_page), info);

	(void)gtk_print_operation_run(print, GTK_PRINT_OPERATION_ACTION_PRINT_DIALOG, GTK_WINDOW(info->window), NULL);

	g_object_unref(print);
}

static char *save_file(const char *title, const char *suggested_filename, GtkWidget * parent, int *success)
{
	GtkWidget *dialog;
	char *filename = NULL;

	*success = 0;
	dialog = gtk_file_chooser_dialog_new(title, GTK_WINDOW(parent), GTK_FILE_CHOOSER_ACTION_SAVE, _("_Cancel"), GTK_RESPONSE_CANCEL, _("_Save"), GTK_RESPONSE_ACCEPT, NULL);

	if (suggested_filename && *suggested_filename) {
		gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(dialog), suggested_filename);
	}

	if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
		filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
		*success = 1;
	}

	gtk_widget_destroy(dialog);

	return filename;
}

static void save_pixbuf(GObject * ignored, gpointer user_data)
{
	ViewerCbInfo *info = (ViewerCbInfo *) user_data;
	char *filename, *base_name, *filename_suggestion;
	int success = 0;
	cairo_surface_t *surface;

	base_name = _rsvg_basename(info->base_uri);
	if (base_name)
		filename_suggestion = g_strdup_printf("%s.png", base_name);
	else
		filename_suggestion = NULL;

	filename = save_file(_("Save SVG as PNG"), filename_suggestion, info->window, &success);
	g_free(base_name);
	g_free(filename_suggestion);

	if (filename) {
		surface = rsvg_image_get_surface(info->image);
		if (cairo_surface_write_to_png(surface, filename) != CAIRO_STATUS_SUCCESS) {
			GtkWidget *errmsg;

			errmsg = gtk_message_dialog_new(GTK_WINDOW(info->window), GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_WARNING, GTK_BUTTONS_CLOSE, "Failed to save");

			gtk_dialog_run(GTK_DIALOG(errmsg));

			gtk_widget_destroy(errmsg);
		}

		g_free(filename);
	}
	else if (success) {
		GtkWidget *errmsg;

		errmsg = gtk_message_dialog_new(GTK_WINDOW(info->window), GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_WARNING, GTK_BUTTONS_CLOSE, _("No filename given"));
		gtk_window_set_transient_for(GTK_WINDOW(errmsg), GTK_WINDOW(info->window));

		gtk_dialog_run(GTK_DIALOG(errmsg));
		gtk_widget_destroy(errmsg);
	}
}

static void quit(GObject *ignored, ViewerCbInfo *info)
{
	gtk_main_quit();
}

static void reload(GObject * ignored, gpointer user_data);

static void copy_svg_location(GObject * ignored, gpointer user_data)
{
	ViewerCbInfo *info = (ViewerCbInfo *) user_data;
	GtkClipboard *clipboard = NULL;

	if (info->base_uri) {
		clipboard = gtk_clipboard_get(GDK_SELECTION_CLIPBOARD);
		gtk_clipboard_set_text(clipboard, info->base_uri, -1);
	}
}

static void create_popup_menu(ViewerCbInfo * info)
{
	GtkWidget *popup_menu;
	GtkWidget *menu_item;

	popup_menu = gtk_menu_new();
	gtk_menu_set_accel_group(GTK_MENU(popup_menu), info->accel_group);

	if (info->base_uri) {
		menu_item = gtk_menu_item_new_with_label(_("Reload"));
		g_signal_connect(menu_item, "activate", G_CALLBACK(reload), info);
		gtk_widget_show(menu_item);
		gtk_menu_shell_append(GTK_MENU_SHELL(popup_menu), menu_item);
		gtk_widget_add_accelerator(menu_item, "activate", info->accel_group, GDK_KEY_R, 0, GTK_ACCEL_VISIBLE);
	}

	if (info->base_uri) {
		menu_item = gtk_menu_item_new_with_label(_("Copy SVG location"));
		g_signal_connect(menu_item, "activate", G_CALLBACK(copy_svg_location), info);
		gtk_widget_show(menu_item);
		gtk_menu_shell_append(GTK_MENU_SHELL(popup_menu), menu_item);
		gtk_widget_add_accelerator(menu_item, "activate", info->accel_group, GDK_KEY_C, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
	}

	menu_item = gtk_menu_item_new_with_label(_("Save as PNG"));
	g_signal_connect(menu_item, "activate", G_CALLBACK(save_pixbuf), info);
	gtk_widget_show(menu_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(popup_menu), menu_item);
	gtk_widget_add_accelerator(menu_item, "activate", info->accel_group, GDK_KEY_S, GDK_CONTROL_MASK | GDK_SHIFT_MASK, GTK_ACCEL_VISIBLE);

	menu_item = gtk_menu_item_new_with_label(_("Print"));
	g_signal_connect(menu_item, "activate", G_CALLBACK(print_pixbuf), info);
	gtk_widget_show(menu_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(popup_menu), menu_item);
	gtk_widget_add_accelerator(menu_item, "activate", info->accel_group, GDK_KEY_P, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);

	menu_item = gtk_menu_item_new_with_label(_("Zoom In"));
	g_signal_connect(menu_item, "activate", G_CALLBACK(zoom_in), info);
	gtk_widget_show(menu_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(popup_menu), menu_item);
	gtk_widget_add_accelerator(menu_item, "activate", info->accel_group, GDK_KEY_plus, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);

	menu_item = gtk_menu_item_new_with_label(_("Zoom Out"));
	g_signal_connect(menu_item, "activate", G_CALLBACK(zoom_out), info);
	gtk_widget_show(menu_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(popup_menu), menu_item);
	gtk_widget_add_accelerator(menu_item, "activate", info->accel_group, GDK_KEY_minus, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);

	menu_item = gtk_menu_item_new_with_label(_("Zoom Extents"));
	g_signal_connect(menu_item, "activate", G_CALLBACK(zoom_extents), info);
	gtk_widget_show(menu_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(popup_menu), menu_item);
	gtk_widget_add_accelerator(menu_item, "activate", info->accel_group, GDK_KEY_e, 0, GTK_ACCEL_VISIBLE);

	menu_item = gtk_menu_item_new_with_label(_("Center"));
	g_signal_connect(menu_item, "activate", G_CALLBACK(center), info);
	gtk_widget_show(menu_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(popup_menu), menu_item);
	gtk_widget_add_accelerator(menu_item, "activate", info->accel_group, GDK_KEY_c, 0, GTK_ACCEL_VISIBLE);

	menu_item = gtk_menu_item_new_with_label(_("Invert background color"));
	g_signal_connect(menu_item, "activate", G_CALLBACK(invert_bg_color), info);
	gtk_widget_show(menu_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(popup_menu), menu_item);
	gtk_widget_add_accelerator(menu_item, "activate", info->accel_group, GDK_KEY_b, 0, GTK_ACCEL_VISIBLE);

	menu_item = gtk_menu_item_new_with_label(_("Reset zoom and pan"));
	g_signal_connect(menu_item, "activate", G_CALLBACK(reset_zoom), info);
	gtk_widget_show(menu_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(popup_menu), menu_item);
	gtk_widget_add_accelerator(menu_item, "activate", info->accel_group, GDK_KEY_z, 0, GTK_ACCEL_VISIBLE);

	menu_item = gtk_menu_item_new_with_label(_("Quit"));
	g_signal_connect(menu_item, "activate", G_CALLBACK(quit), info);
	gtk_widget_show(menu_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(popup_menu), menu_item);
	gtk_widget_add_accelerator(menu_item, "activate", info->accel_group, GDK_KEY_q, 0, GTK_ACCEL_VISIBLE);

	info->popup_menu = popup_menu;
}


static gint button_event(GObject * widget, GdkEventButton * event, gpointer user_data)
{
	ViewerCbInfo *info = (ViewerCbInfo *)user_data;

	if (event->button == 3 && event->type == GDK_BUTTON_PRESS) {
		gtk_menu_popup(GTK_MENU(info->popup_menu), NULL, NULL, NULL, NULL, event->button, event->time);
		return TRUE;
	}

	if ((event->button == 1) && (event->type == GDK_BUTTON_PRESS)) {
		printf("SVG coord: %f;%f\n", ATX(info, event->x), ATY(info, event->y));
		fflush(stdout);
		return TRUE;
	}

	if ((event->button == 2) && (event->type == GDK_BUTTON_PRESS)) {
		info->pan_aox = info->x_pan;
		info->pan_aoy = info->y_pan;
		info->pan_mox = event->x;
		info->pan_moy = event->y;
		info->panning = 1;
		return TRUE;
	}

	if ((event->button == 2) && (event->type == GDK_BUTTON_RELEASE)) {
		info->panning = 0;
		return TRUE;
	}


	return FALSE;
}

static gint motion_event(GObject * widget, GdkEventButton * event, gpointer user_data)
{
	ViewerCbInfo *info = (ViewerCbInfo *) user_data;

	/* need to remember last mouse coords for smart zoom */
	info->mouse_x = event->x;
	info->mouse_y = event->y;

	if (!info->panning)
		return FALSE;

	info->x_pan = info->pan_aox + (event->x - info->pan_mox);
	info->y_pan = info->pan_aoy + (event->y - info->pan_moy);

	rsvg_image_take_surface(info->image, render_to_surface(info));

	return TRUE;
}

static gint scroll_event(GObject * widget, GdkEventScroll * event, gpointer user_data)
{
	if (event->direction == GDK_SCROLL_UP) zoom_image(user_data, 1.2);
	else if (event->direction == GDK_SCROLL_DOWN) zoom_image(user_data, 0.8);
	return TRUE;
}

static gint key_event(GObject * widget, GdkEventKey * event, gpointer user_data)
{
	switch(event->keyval) {
		case GDK_KEY_plus:
		case GDK_KEY_KP_Add:
			zoom_image(user_data, 1.2);
			return TRUE;

		case GDK_KEY_minus:
		case GDK_KEY_KP_Subtract:
			zoom_image(user_data, 0.8);
			return TRUE;

		case GDK_KEY_Left:
		case GDK_KEY_KP_Left:
			pan_image(user_data, +100, 0);
			return TRUE;

		case GDK_KEY_Right:
		case GDK_KEY_KP_Right:
			pan_image(user_data, -100, 0);
			return TRUE;

		case GDK_KEY_Up:
		case GDK_KEY_KP_Up:
			pan_image(user_data, 0, +100);
			return TRUE;

		case GDK_KEY_Down:
		case GDK_KEY_KP_Down:
			pan_image(user_data, 0, -100);
			return TRUE;
	}
	return FALSE;
}


static void quit_cb(GtkWidget * win, gpointer unused)
{
	/* exit the main loop */
	gtk_main_quit();
}

static GtkToolItem *tool_button_new(const char *icon_name)
{
	GtkToolItem *toolitem;

	toolitem = gtk_tool_button_new(NULL, NULL);
	gtk_tool_button_set_icon_name(GTK_TOOL_BUTTON(toolitem), icon_name);
	return toolitem;
}

static gint resized(GObject * widget, GdkEvent *ev, gpointer user_data)
{
	ViewerCbInfo *info = (ViewerCbInfo *) user_data;
	rsvg_image_take_surface(info->image, render_to_surface(info));
	return FALSE;
}


static gboolean want_toolbar = 0;

static void drawing_add(ViewerCbInfo *info, cairo_surface_t * surface)
{
	/* create a new image */
	info->last_surface = surface;
	info->image = rsvg_image_new_take_surface(surface);
	gtk_box_pack_start(GTK_BOX(info->drawing_host), GTK_WIDGET(info->image), TRUE, TRUE, 0);

	info->image_valid = 1;
}

static void drawing_del(ViewerCbInfo *info)
{
	info->image_valid = 0;
	gtk_widget_destroy(GTK_WIDGET(info->image));
}


static void reload(GObject * ignored, gpointer user_data)
{
	ViewerCbInfo *info = (ViewerCbInfo *) user_data;
	cairo_surface_t *surface = info->last_surface;
	gdouble px, py, zx, zy;

	/* save zoom/pan */
	px = info->x_pan;
	py = info->y_pan;
	zx = info->x_zoom;
	zy = info->y_zoom;

	/* remove the old widget */
	drawing_del(info);
	g_object_unref(info->handle);

	/* load and creater a new one */
	surface = load(1);
	drawing_add(info, surface);

	/* restore zoom/pan */
	info->x_pan = px;
	info->y_pan = py;
	info->x_zoom = zx;
	info->y_zoom = zy;
	gtk_widget_show_all(info->drawing_host);
	rsvg_image_take_surface(info->image, render_to_surface(info));

	/* window title shows the reload counter so the user knows */
	reload_ctr++;
	set_window_title(info);
}


static void populate_window(GtkWidget * win, ViewerCbInfo * info, cairo_surface_t * surface /* adopted */ )
{
	info->drawing_host = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
	gtk_container_add(GTK_CONTAINER(win), info->drawing_host);

	if (want_toolbar) {
		GtkWidget *toolbar;
		GtkToolItem *toolitem;

		toolbar = gtk_toolbar_new();
		gtk_box_pack_start(GTK_BOX(info->drawing_host), toolbar, FALSE, FALSE, 0);

		toolitem = tool_button_new("zoom-in");
		gtk_toolbar_insert(GTK_TOOLBAR(toolbar), toolitem, 0);
		g_signal_connect(toolitem, "clicked", G_CALLBACK(zoom_in), info);

		toolitem = tool_button_new("zoom-out");
		gtk_toolbar_insert(GTK_TOOLBAR(toolbar), toolitem, 1);
		g_signal_connect(toolitem, "clicked", G_CALLBACK(zoom_out), info);
	}
	drawing_add(info, surface);
}

static void view_surface(ViewerCbInfo * info, cairo_surface_t * surface /* adopted */ ,
												 const char *bg_color_str)
{
	GtkWidget *win;

	/* create toplevel window and set its title */

	win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_default_size(win, 640, 480);

	populate_window(win, info, surface);

	/* exit when 'X' is clicked */
	g_signal_connect(win, "destroy", G_CALLBACK(quit_cb), NULL);
	g_signal_connect(win, "delete_event", G_CALLBACK(quit_cb), NULL);

	gdk_rgba_parse(&info->bg_color, "#ffffff");
	if ((bg_color_str != NULL) && strcmp(bg_color_str, "none") != 0) {
		if (gdk_rgba_parse(&info->bg_color, bg_color_str)) {
			GtkWidget *parent_widget = gtk_widget_get_parent(GTK_WIDGET(info->image));
			gtk_widget_override_background_color(parent_widget, GTK_STATE_FLAG_NORMAL, &info->bg_color);
		}
		else
			g_warning(_("Couldn't parse color '%s'"), bg_color_str);
	}

	create_popup_menu(info);

	info->window = win;
	gtk_window_add_accel_group(GTK_WINDOW(win), info->accel_group);

	gtk_widget_add_events(win, GDK_SCROLL_MASK);
	gtk_widget_add_events(win, GDK_POINTER_MOTION_MASK);

	g_signal_connect(win, "button-press-event", G_CALLBACK(button_event), info);
	g_signal_connect(win, "button-release-event", G_CALLBACK(button_event), info);
	g_signal_connect(win, "scroll-event", G_CALLBACK(scroll_event), info);
	g_signal_connect(win, "motion-notify-event", G_CALLBACK(motion_event), info);
	g_signal_connect(win, "key-release-event", G_CALLBACK(key_event), info);
	g_signal_connect(win, "configure-event", G_CALLBACK(resized), info);

	gtk_widget_show_all(win);

	set_window_title(info);
}


/* Original settings and file load state thta need to be persisten between main
   and subsequent invocations of load() */
static char **args = NULL;
static gint n_args = 0;
static char *base_uri = NULL;
static ViewerCbInfo info;
static char *id = NULL;
static double x_zoom = 1.0;
static double y_zoom = 1.0;
static double dpi_x = -1.0;
static double dpi_y = -1.0;
static RsvgHandleFlags flags = RSVG_HANDLE_FLAGS_NONE;
static int width = -1;
static int height = -1;

static cairo_surface_t *load(int reload)
{
	GInputStream *input;
	GError *err = NULL;
	GFile *file, *base_file;
	GFileInfo *file_info;
	gboolean compressed = FALSE;
	cairo_surface_t *surface;

	{ /* load the file */
		file = g_file_new_for_commandline_arg(args[0]);
		input = (GInputStream *) g_file_read(file, NULL, &err);

		if (base_uri)
			base_file = g_file_new_for_uri(base_uri);
		else
			base_file = g_object_ref(file);

		if ((file_info = g_file_query_info(file, G_FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE, G_FILE_QUERY_INFO_NONE, NULL, NULL))) {
			const char *content_type;
			char *gz_content_type;

			content_type = g_file_info_get_content_type(file_info);
			gz_content_type = g_content_type_from_mime_type("application/x-gzip");
			compressed = (content_type != NULL && g_content_type_is_a(content_type, gz_content_type));
			g_free(gz_content_type);
			g_object_unref(file_info);
		}

		g_object_unref(file);
	}

	if (input == NULL) {
		g_printerr("Failed to read input: %s\n", err->message);
		g_error_free(err);
		return NULL;
	}

	if (compressed) {
		GZlibDecompressor *decompressor;
		GInputStream *converter_stream;

		decompressor = g_zlib_decompressor_new(G_ZLIB_COMPRESSOR_FORMAT_GZIP);
		converter_stream = g_converter_input_stream_new(input, G_CONVERTER(decompressor));
		g_object_unref(input);
		input = converter_stream;
	}

	info.base_uri = base_file ? g_file_get_uri(base_file) : g_strdup("");
	info.id = id;
	info.x_zoom = x_zoom;
	info.y_zoom = y_zoom;

	info.handle = rsvg_handle_new_from_stream_sync(input, base_file, flags, NULL /* cancellable */ ,
																								 &err);
	if (base_file != NULL)
		g_object_unref(base_file);
	g_object_unref(input);

	if (info.handle == NULL) {
		g_printerr("Failed to load SVG: %s\n", err->message);
		g_error_free(err);
		return NULL;
	}

	rsvg_handle_get_dimensions(info.handle, &info.dimensions);

	if (width != -1) {
		info.x_zoom = (double)width / info.dimensions.width;
	}
	else {
		info.x_zoom = x_zoom;
	}
	if (height != -1) {
		info.y_zoom = (double)height / info.dimensions.height;
	}
	else {
		info.y_zoom = y_zoom;
	}

	surface = render_to_surface(&info);
	if (surface == NULL) {
		g_printerr("Unknown error while rendering image\n");
		return NULL;
	}

	return surface;
}

int main(int argc, char **argv)
{
	int retval = 1;
	GOptionContext *g_option_context;
	int bVersion = 0;
	char *bg_color = NULL;
	gboolean keep_aspect_ratio = FALSE;
	gboolean unlimited = FALSE;
	cairo_surface_t *surface;

	GOptionEntry options_table[] = {
		{"toolbar", 't', 0, G_OPTION_ARG_NONE, &want_toolbar, N_("Display toolbar"),
		 N_("<bool>")},
		{"dpi-x", 'd', 0, G_OPTION_ARG_DOUBLE, &dpi_x, N_("Set the # of Pixels Per Inch"),
		 N_("<float>")},
		{"dpi-y", 'p', 0, G_OPTION_ARG_DOUBLE, &dpi_y, N_("Set the # of Pixels Per Inch"),
		 N_("<float>")},
		{"x-zoom", 'x', 0, G_OPTION_ARG_DOUBLE, &x_zoom, N_("Set the x zoom factor"),
		 N_("<float>")},
		{"y-zoom", 'y', 0, G_OPTION_ARG_DOUBLE, &y_zoom, N_("Set the y zoom factor"),
		 N_("<float>")},
		{"width", 'w', 0, G_OPTION_ARG_INT, &width, N_("Set the image's width"), N_("<int>")},
		{"height", 'h', 0, G_OPTION_ARG_INT, &height, N_("Set the image's height"), N_("<int>")},
		{"bg-color", 'b', 0, G_OPTION_ARG_STRING, &bg_color,
		 N_("Set the image background color (default: transparent)"), N_("<string>")},
		{"base-uri", 'u', 0, G_OPTION_ARG_STRING, &base_uri, N_("Set the base URI (default: none)"),
		 N_("<string>")},
		{"id", 0, 0, G_OPTION_ARG_STRING, &id, N_("Only show one node (default: all)"),
		 N_("<string>")},
		{"keep-aspect", 'k', 0, G_OPTION_ARG_NONE, &keep_aspect_ratio,
		 N_("Preserve the image's aspect ratio"), NULL},
		{"unlimited", 'u', 0, G_OPTION_ARG_NONE, &unlimited,
		 N_("Allow huge SVG files"), NULL},
		{"version", 'v', 0, G_OPTION_ARG_NONE, &bVersion, N_("Show version information"), NULL},
		{G_OPTION_REMAINING, 0, 0, G_OPTION_ARG_FILENAME_ARRAY, &args, NULL, N_("[FILE...]")},
		{NULL}
	};

	/* Set the locale so that UTF-8 filenames work */
	setlocale(LC_ALL, "");

	RSVG_G_TYPE_INIT;

	info.window = NULL;
	info.popup_menu = NULL;

	g_option_context = g_option_context_new("- SVG Viewer");
	g_option_context_add_main_entries(g_option_context, options_table, NULL);
	g_option_context_add_group(g_option_context, gtk_get_option_group(TRUE));
	g_option_context_set_help_enabled(g_option_context, TRUE);
	if (!g_option_context_parse(g_option_context, &argc, &argv, NULL)) {
		exit(1);
	}

	g_option_context_free(g_option_context);

	if (bVersion != 0) {
		g_message(_("rsvg-view version %s\n"), LIBRSVG_VERSION);
		return 0;
	}

	if (args)
		n_args = g_strv_length(args);
	else
		n_args = 0;

	if (n_args != 1) {
		g_printerr(_("Exactly one file name needs to be specified\n"));
		return 1;
	}

	rsvg_set_default_dpi_x_y(dpi_x, dpi_y);


	if (unlimited)
		flags |= RSVG_HANDLE_FLAG_UNLIMITED;

	surface = load(0);
	if (surface == NULL)
		goto done;

	retval = 0;

	info.accel_group = gtk_accel_group_new();

	view_surface(&info, surface, bg_color);

	/* run the gtk+ main loop */
	gtk_main();

done:

	g_strfreev(args);
	g_free(info.base_uri);
	g_object_unref(info.handle);

	rsvg_cleanup();

	return retval;
}
